////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//   Wireless Communication Library 7                                         //
//                                                                            //
//   Copyright (C) 2006-2024 Mike Petrichenko                                 //
//                           Soft Service Company                             //
//                           All Rights Reserved                              //
//                                                                            //
//   http://www.btframework.com                                               //
//                                                                            //
//   support@btframework.com                                                  //
//   shop@btframework.com                                                     //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include "..\Common\wclErrors.h"

#include "..\Communication\wclConnections.h"

#include "wclIrDADevices.h"

using namespace wclCommon;
using namespace wclCommunication;

namespace wclIrDA
{
	/// <summary> The IrDA client component. </summary>
	/// <remarks> The IrDA Client component allows your application to connect
	///   to a remote IrDA enabled device and exchange data with the
	///   device. </remarks>
	class CwclIrDAClient
	{
		DISABLE_COPY(CwclIrDAClient);

	private:
		unsigned long				FAddress;
		wclIrDAConnectionMode		FMode;
		tstring						FService;
		unsigned long				FTimeout;

		CwclClientDataConnection*	FConnection;

		void ClientConnect(void* Sender, const int Error);
		void ClientData(void* Sender, const void* const Data, const unsigned long Size);
		void ClientDisconnect(void* Sender, const int Reason);
		
	protected:
		/// <summary> Fires the <c>OnConnect</c> event. </summary>
		/// <param name="Error"> The connection operation result code. If the
		///   parameter is <see cref="WCL_E_SUCCESS" /> the connection to a remote
		///   device was established and the remote device is connected. Otherwise
		///   the connection was not established and the remote device is
		///   not connected. </param>
		virtual void DoConnect(const int Error);
		/// <summary> Fires the <c>OnCreateProcessor</c> event. </summary>
		/// <param name="Connection"> The <see cref="CwclClientDataConnection" />
		///   object that requires a data processor. </param>
		/// <seealso cref="CwclClientDataConnection" />
		virtual void DoCreateProcessor(CwclClientDataConnection* const Connection);
		/// <summary> Fires the <c>OnData</c> event. </summary>
		/// <param name="Data"> The pointer to the received data buffer.
		///   It is guaranteed that the parameter points to a valid data buffer.
		///   The data buffer is valid only inside the event handler. If an
		///   application needs to use the data outside the event handle it must
		///   allocate own buffer and copy data into it. </param>
		/// <param name="Size"> The data buffer size. It is guaranteed that the size
		///   is greater than 0. </param>
		virtual void DoData(const void* const Data, const unsigned long Size);
		/// <summary> Fires the <c>OnDestroyProcessor</c> event. </summary>
		/// <param name="Connection"> The <see cref="CwclClientDataConnection" />
		///   object that requires a data processor. </param>
		/// <seealso cref="CwclClientDataConnection" />
		virtual void DoDestroyProcessor(CwclClientDataConnection* const Connection);
		/// <summary> Fires the <c>OnDisconnect</c> event. </summary>
		/// <param name="Reason"> The disconnection code. If the
		///   parameter is <see cref="WCL_E_SUCCESS" /> the connection was closed
		///   by the <c>Disconnect</c> call. Otherwise indicates the disconnection
		///   reason. </param>
		virtual void DoDisconnect(const int Reason);

	public:
		/// <summary> Create a new IrDA client. </summary>
		CwclIrDAClient();
		/// <summary> Frees the component. </summary>
		virtual ~CwclIrDAClient();

		/// <summary> Connects to a remote device. </summary>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int Connect();
		/// <summary> Disconnects from the connected device. </summary>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int Disconnect();

		/// <summary> Reads the read buffer size. </summary>
		/// <param name="Size"> On output he read buffer size in bytes. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int GetReadBufferSize(unsigned long& Size);
		/// <summary> Reads the write buffer size. </summary>
		/// <param name="Size"> On output the write buffer size in bytes. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int GetWriteBufferSize(unsigned long& Size);

		/// <summary> Sets the read buffer size. </summary>
		/// <param name="Size"> The new read buffer size in bytes. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int SetReadBufferSize(const unsigned long Size);
		/// <summary> Sets the write buffer size. </summary>
		/// <param name="Size"> The new write buffer size in bytes. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int SetWriteBufferSize(const unsigned long Size);

		/// <summary> Sends data to the connected device. </summary>
		/// <param name="Data"> The pointer to the data buffer that should be sent
		///   to the connected device. </param>
		/// <param name="Size"> The data buffer size. </param>
		/// <param name="Written"> The amount of bytes actually written to the
		///   device. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		/// <remarks> The write is synchronous operation. </remarks>
		int Write(const void* const Data, const unsigned long Size,
			unsigned long& Written);

		/// <summary> Gets the Data Processor created for the connection. </summary>
		/// <returns> The <see cref="CwclCustomClientDataProcessor" />
		///   object. </returns>
		/// <seealso cref="CwclCustomClientDataProcessor" />
		CwclCustomClientDataProcessor* GetProcessor() const;
		/// <summary> Gets the Data Processor created for the connection. </summary>
		/// <value> The <see cref="CwclCustomClientDataProcessor" />
		///   object. </value>
		/// <seealso cref="CwclCustomClientDataProcessor" />
		__declspec(property(get = GetProcessor)) CwclCustomClientDataProcessor* Processor;

		/// <summary> Gets the connection state. </summary>
		/// <returns> The <see cref="wclClientState" /> indicates the current
		///   connection state. </returns>
		/// <seealso cref="wclClientState" />
		wclClientState GetState() const;
		/// <summary> Gets the connection state. </summary>
		/// <value> The <see cref="wclClientState" /> indicates the current
		///   connection state. </value>
		/// <seealso cref="wclClientState" />
		__declspec(property(get = GetState)) wclClientState State;

		/// <summary> Gets the target IrDA device's address. </summary>
		/// <returns> The remote IrDA device's MAC address. </returns>
		unsigned long GetAddress() const;
		/// <summary> Sets the target IrDA device's address. </summary>
		/// <param name="Value"> The remote IrDA device's MAC address. </param>
		/// <remarks> If an application changes the property when client is
		///   connected the <see cref="wclEClientActive" /> exception
		///   raises. </remarks>
		/// <exception cref="wclEClientActive"></exception>
		void SetAddress(const unsigned long Value);
		/// <summary> Gets and sets the target IrDA device's address. </summary>
		/// <value> The remote IrDA device's MAC address. </value>
		/// <remarks> If an application changes the property when client is
		///   connected the <see cref="wclEClientActive" /> exception
		///   raises. </remarks>
		/// <exception cref="wclEClientActive"></exception>
		__declspec(property(get = GetAddress, put = SetAddress)) unsigned long Address;

		/// <summary> Gets the IrDA connection mode. </summary>
		/// <returns> The IrDA connection mode. </returns>
		/// <seealso cref="wclIrDAConnectionMode"/>
		wclIrDAConnectionMode GetMode() const;
		/// <summary> Sets the IrDA connection mode. </summary>
		/// <param name="Value"> The IrDA connection mode. </param>
		/// <remarks> If an application changes the property when client is
		///   connected the <see cref="wclEClientActive" /> exception
		///   raises. </remarks>
		/// <seealso cref="wclIrDAConnectionMode"/>
		/// <exception cref="wclEClientActive"></exception>
		void SetMode(const wclIrDAConnectionMode Value = cmIrComm9Wire);
		/// <summary> Gets and sets the IrDA connection mode. </summary>
		/// <value> The IrDA connection mode. </value>
		/// <remarks> If an application changes the property when client is
		///   connected the <see cref="wclEClientActive" /> exception
		///   raises. </remarks>
		/// <seealso cref="wclIrDAConnectionMode" />
		/// <exception cref="wclEClientActive"></exception>
		__declspec(property(get = GetMode, put = SetMode)) wclIrDAConnectionMode Mode;

		/// <summary> Gets the IrDA service that should be used to
		///   connect. It can be a TinyTP service name or LSAP selector number.
		///   In case of using LSAP selector number the format is 'LSAP-SELn'
		///   where <c>n</c> is a select number in range from 1 to 127. </summary>
		/// <returns> The IrDA service's name or LSAP selector number. </returns>
		tstring GetService() const;
		/// <summary> Sets the IrDA service that should be used to
		///   connect. It can be a TinyTP service name or LSAP selector number.
		///   In case of using LSAP selector number the format is 'LSAP-SELn'
		///   where <c>n</c> is a select number in range from 1 to 127. </summary>
		/// <param name="Value"> The IrDA service's name or LSAP selector number. </param>
		/// <remarks> If an application changes the property when client is
		///   connected the <see cref="wclEClientActive" /> exception
		///   raises. If new value of the property is an empty string the
		///   <see cref="wclEInvalidArgument" /> exception raises. </remarks>
		/// <exception cref="wclEClientActive"></exception>
		/// <exception cref="wclEInvalidArgument"></exception>
		void SetService(const tstring& Value = _T("IrDA:IrCOMM"));
		/// <summary> Gets and sets the IrDA service that should be used to
		///   connect. It can be a TinyTP service name or LSAP selector number.
		///   In case of using LSAP selector number the format is 'LSAP-SELn'
		///   where <c>n</c> is a select number in range from 1 to 127. </summary>
		/// <value> The IrDA service's name or LSAP selector number. </value>
		/// <remarks> If an application changes the property when client is
		///   connected the <see cref="wclEClientActive" /> exception
		///   raises. If new value of the property is an empty string the
		///   <see cref="wclEInvalidArgument" /> exception raises. </remarks>
		/// <exception cref="wclEClientActive"></exception>
		/// <exception cref="wclEInvalidArgument"></exception>
		__declspec(property(get = GetService, put = SetService)) tstring Service;

		/// <summary> Gets the connect timeout value. </summary>
		/// <returns> The connect timeout value in milliseconds. </returns>
		unsigned long GetTimeout() const;
		/// <summary> Sets connect timeout value. </summary>
		/// <param name="Value"> The connect timeout value in milliseconds. </param>
		/// <remarks> If an application changes the property when client is
		///   connected the <see cref="wclEClientActive" /> exception
		///   raises. If new value of the <c>Timeout</c> is <c>0</c> the
		///   <see cref="wclEInvalidArgument" /> exception raises. </remarks>
		/// <exception cref="wclEClientActive"></exception>
		/// <exception cref="wclEInvalidArgument"></exception>
		void SetTimeout(const unsigned long Value = 10000);
		/// <summary> Gets and sets the connect timeout value. </summary>
		/// <value> The connect timeout value in milliseconds. </value>
		/// <remarks> If an application changes the property when client is
		///   connected the <see cref="wclEClientActive" /> exception
		///   raises. If new value of the <c>Timeout</c> is <c>0</c> the
		///   <see cref="wclEInvalidArgument" /> exception raises. </remarks>
		/// <exception cref="wclEClientActive"></exception>
		/// <exception cref="wclEInvalidArgument"></exception>
		__declspec(property(get = GetTimeout, put = SetTimeout)) unsigned long Timeout;

		/// <summary> The event fires when a connection to a remote device
		///   has been completed (with or without success). </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Error"> The connection operation result code. If the
		///   parameter is <see cref="WCL_E_SUCCESS" /> the connection to a remote
		///   device was established and the remote device is connected. Otherwise
		///   the connection was not established and the remote device is
		///   not connected. </param>
		/// <remarks> If the <c>Error</c> parameter of the event is not
		///   <see cref="WCL_E_SUCCESS" /> an application may call <c>Connect</c>
		///   method again from inside the event handler. </remarks>
		wclClientConnectionConnectEvent(OnConnect);
		/// <summary> The event fires when connection to a remote device has
		///   been established with success to allow application provide a
		///   data processor for the connection. </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Connection"> The <see cref="CwclClientDataConnection" />
		///   object that requires a Data Processor. </param>
		/// <remarks> If an application needs an additional data processing it
		///   can use Data Processors. An application must create a Data Processor
		///   inside this event handler. If a Data Processor created for the
		///   connection the <c>OnData</c> event will NOT fire. </remarks>
		/// <seealso cref="CwclClientDataConnection" />
		/// <seealso cref="CwclCustomClientDataProcessor" />
		wclClientConnectionProcessorEvent(OnCreateProcessor);
		/// <summary> The event fires when new data is received from the
		///   connected remote device. </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Data"> The pointer to the received data buffer.
		///   It is guaranteed that the parameter points to a valid data buffer.
		///   The data buffer is valid only inside the event handler. If an
		///   application needs to use the data outside the event handle it must
		///   allocate own buffer and copy data into it. </param>
		/// <param name="Size"> The data buffer size. It is guaranteed that the size
		///   is greater than 0. </param>
		wclConnectionDataEvent(OnData);
		/// <summary> The event fires when connection to a remote device has
		///   been terminated. An application must destroy the Data Processor
		///   created for the connection. </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Connection"> The <see cref="CwclClientDataConnection" />
		///   object that requires a Data Processor. </param>
		/// <seealso cref="CwclClientDataConnection" />
		/// <seealso cref="CwclCustomClientDataProcessor" />
		wclClientConnectionProcessorEvent(OnDestroyProcessor);
		/// <summary> The event fires when the remote connected device has been
		///   disconnected. </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Reason"> The disconnection code. If the
		///   parameter is <see cref="WCL_E_SUCCESS" /> the connection was closed
		///   by the <c>Disconnect</c> call. Otherwise indicates the disconnection
		///   reason. </param>
		wclClientConnectionDisconnectEvent(OnDisconnect);
	};
}