////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//   Wireless Communication Library 7                                         //
//                                                                            //
//   Copyright (C) 2006-2024 Mike Petrichenko                                 //
//                           Soft Service Company                             //
//                           All Rights Reserved                              //
//                                                                            //
//   http://www.btframework.com                                               //
//                                                                            //
//   support@btframework.com                                                  //
//   shop@btframework.com                                                     //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include <vector>
#include <set>

#include <winsock2.h>

#include "..\Common\wclHelpers.h"
#include "..\Common\wclMessaging.h"

#include "wclIrDAErrors.h"

using namespace wclCommon;

namespace wclIrDA
{
	/// <summary> The IrDA connection modes. </summary>
	typedef enum
	{
		/// <summary> The IrDA TinyTP 3-wire connection mode. </summary>
		cmIrComm3Wire,
		/// <summary> The IrDA TinyTP 9-wire connection mode (also known as
		///   "Cooked Mode"). This option causes a connection to use cooked mode on
		///   the data transferred.  Note that no access to the received control
		///   information is available in Windows and null control information is
		///   sent. Note also that it is not possible to neither manually add
		///   control information alongside outgoing data nor intercept incoming
		///   control information. </summary>
		cmIrComm9Wire,
		/// <summary> IrLMP (non-TinyTP) mode. Note that this mode is not supported
		///   on a server side. </summary>
		cmIrLpt
	} wclIrDAConnectionMode;

	/// <summary> IrDA character sets. </summary>
	typedef enum
	{
		/// <summary> ASCII character set. </summary>
		csAscii,
		/// <summary> ISO 8859-1 (Latin-1, Western European) character set. </summary>
		csIso8859_1,
		/// <summary> ISO 8859-2 (Latin-2, Central European) character set. </summary>
		csIso8859_2,
		/// <summary> ISO 8859-3 (Latin-3, South European) character set. </summary>
		csIso8859_3,
		/// <summary> ISO 8859-4 (Latin-4, North European) character set. </summary>
		csIso8859_4,
		/// <summary> ISO 8859-5 (Latin/Cyrillic) character set. </summary>
		csIso8859_5,
		/// <summary> ISO 8859-6 (Latin/Arabic) character set. </summary>
		csIso8859_6,
		/// <summary> ISO 8859-7 (Latin/Greek) character set. </summary>
		csIso8859_7,
		/// <summary> ISO 8859-8 (Latin/Hebrew) character set. </summary>
		csIso8859_8,
		/// <summary> ISO 8859-9 (Latin-5, Turkish) character set. </summary>
		csIso8859_9,
		/// <summary> UNICODE character set. </summary>
		csUnicode
	} wclIrDACharSet;

	/// <summary> The hint can give information about what a given IrDA device is
	///   and what it supports. </summary>
	typedef enum
	{
		/// <summary> The device supports PnP. </summary>
		ihPnp,
		/// <summary> The device is a PDA. </summary>
		ihPda,
		/// <summary> The device is a computer. </summary>
		ihComputer,
		/// <summary> The device supports printer features. </summary>
		ihPrinter,
		/// <summary> The device supports modem features. </summary>
		ihModem,
		/// <summary> The device supports fax features. </summary>
		ihFax,
		/// <summary> The device supports LAN features. </summary>
		ihLan,
		/// <summary> The device supports telephony features. </summary>
		ihTelephony,
		/// <summary> The device supports file server features. </summary>
		ihFileServer,
		/// <summary> The device supports communication features. </summary>
		ihComm,
		/// <summary> The device supports messaging features. </summary>
		ihMessage,
		/// <summary> The device supports HTTP. </summary>
		ihHttp,
		/// <summary> The device supports OBEX. </summary>
  		 ihObex
	} wclIrDAHint;
	/// <summary> Set of the IrDA hints. </summary>
	/// <seealso cref="wclIrDAHint" />
	typedef std::set<wclIrDAHint> wclIrDAHints;

	/// <summary> This record contains information about an IrDA device that has
	///   been discovered by the infrared transceiver. </summary>
	typedef struct
	{
		/// <summary> The IrDA device's address. </summary>
		/// <remarks> This address is not permanent, however: if a device will leave
		///   the scope of IR receiver for a significant time interval (30 seconds
		///   or so), it will be given a new address when it returns. </remarks>
		unsigned long	Address;
		/// <summary> Name of the IrDA device. </summary>
		/// <remarks> <para> As the device's address is random an application must
		///   use the device's name to identify required IrDA device. </para>
		///   <para> The name is encoded using character set specified in the
		///   <c>CharSet</c> member. </para> </remarks>
		tstring			Name;
		/// <summary> IrDA device hints. </summary>
		/// <remarks> Hints describe a nature of the device and what services it can
		///   be expected to provide. </remarks>
		/// <seealso cref="wclIrDAHints" />
		wclIrDAHints	Hints;
		/// <summary> The IrDA device's character set. </summary>
		/// <seealso cref="wclIrDACharSet" />
		wclIrDACharSet	CharSet;
	} wclIrDADevice;
	/// <summary> The array of discovered IrDA devices. </summary>
	  /// <seealso cref="wclIrDADevice" />
	typedef std::vector<wclIrDADevice> wclIrDADevices;

	/// <summary> IrDA IAS attribute type. </summary>
	typedef enum
	{
		/// <summary> Integer attribute value. </summary>
		atInteger,
		/// <summary> Bytes sequence attribute value. </summary>
		atSequence,
		/// <summary> String attribute value. </summary>
		atString
	} wclIrDAAttributeType;

	/// <summary> The record describes the IrDA IAS attribute value. </summary>
	typedef struct
	{
		/// <summary> The attribute's value type. </summary>
		/// <seealso cref="wclIrDAAttributeType" />
		wclIrDAAttributeType		ValueType;
		/// <summary> Contains the attribute's integer value. The member is valid
		///   only of the <c>ValueType</c> is <c>atInteger</c>. </summary>
		/// <seealso cref="wclIrDAAttributeType" />
		int							IntegerValue;
		/// <summary> Contains the attribute's bytes sequence value. The member is
		///   valid only if the <c>ValueType</c> is <c>atSequence</c>. </summary>
		/// <seealso cref="wclIrDAAttributeType" />
		std::vector<unsigned char>	SequenceValue;
		/// <summary> Contains the attribute's string value. The member is valid
		///   only if the <c>ValueType</c> is <c>atString</c>. The string is
		///   encoded using character set specified in the <c>CharSet</c>
		///   member. </summary>
		/// <seealso cref="wclIrDAAttributeType" />
		/// <seealso cref="wclIrDACharSet" />
		tstring						StringValue;
		/// <summary> The string value character set. The member is valid
		///   only if the <c>ValueType</c> is <c>atString</c>. </summary>
		/// <seealso cref="wclIrDAAttributeType" />
		/// <seealso cref="wclIrDACharSet" />
		wclIrDACharSet				CharSet;
	} wclIrDAAttribute;

	/// <summary> The <c>OnDevicesFound</c> even handler prototype. </summary>
	/// <param name="Sender"> The object initiates the event. </param>
	/// <param name="Device"> The record represents a just found IrDA
	///   device. </param>
	/// <seealso cref="wclIrDADevice" />
	#define wclIrDADevicesFoundEvent(_event_name_) \
		__event void _event_name_(void* Sender, const wclIrDADevice& Device)

	/// <summary> Represents the IrDA receiver. </summary>
	/// <remarks> <para> The component provides methods to discover IrDA devices
	///   and to monitor changes in the IrDA hardware. </para>
	///   <para> The component is not thread-safe and must be used in
	///   the same thread it was created. </para> </remarks>
	class CwclIrDAReceiver
	{
		DISABLE_COPY(CwclIrDAReceiver);

	private:
		bool					FAvailable; // For internal use when subscribed for HW changes.
		RTL_CRITICAL_SECTION	FCS; // Used to access lazy discovering and subscribe methods.
		int						FError; // Lazy discovering thread initialization result.
		HANDLE					FTerminateEvent; // Lazy discovering termination event.
		HANDLE					FInitEvent; // Lazy discovering initialization event.
		CwclMessageReceiver*	FReceiver;
		bool					FSubscribed; // Hardware changes subscription state.
		HANDLE					FThread; // Lazy discovering thread.

		static UINT __stdcall _ThreadProc(LPVOID lpParam);
		void LazyDiscoveringThreadProc();
		
	protected:
		/// <summary> The message receiver calls this method when a new notification
		///   message has been received. </summary>
		/// <param name="Message"> A <see cref="CwclMessage" /> object
		///   represents the notification message. </param>
		/// <remarks> A derived class must always call inherited method in case
		///   it is overrides. </remarks>
		/// <seealso cref="CwclMessage" />
		virtual void MessageReceived(const CwclMessage* const Message);
		
		/// <summary> Fires the <c>OnChange</c> event. </summary>
		virtual void DoChange();
		/// <summary> Fires the <c>OnDevicesFound</c> event. </summary>
		/// <param name="Device"> The record represents the just found
		///   IrDA device. </param>
		/// <seealso cref="wclIrDADevice" />
   		 virtual void DoDevicesFound(const wclIrDADevice& Device);

	public:
		/// <summary> Create a new component. </summary>
		CwclIrDAReceiver();
		/// <summary> Frees the component. </summary>
		virtual ~CwclIrDAReceiver();

		/// <summary> Discovers nearby IrDA devices. </summary>
		/// <param name="Devices"> <para> The array of found IrDA devices. </para>
		///   <para> An application is responsible to release the memory
		///   allocated for the <c>Devices</c> dynamic array. </para> </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		/// <remarks> <para> The method executes discovering on all connected IrDA
		///   adapters. </para>
		///   <para> Even if method succeed the <c>Devices</c> parameter can be
		///   set to <c>NULL</c> if no one IrDA device was found. An application
		///   must always check the returned valid of the <c>Devices</c>
		///   parameter. </para>
		///   <para> This method can be called from separate
		///   thread. </para> </remarks>
		/// <seealso cref="wclIrDADevices" />
		int Discover(wclIrDADevices& Devices);

		/// <summary> Starts Lazy Discovering for IrDA devices. </summary>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		/// <remarks> The Lazy Discovering does not return devices list immediately.
		///   Instead of that it fires the <c>OnDeviceFound</c> event when new
		///   IrDA device appears. </remarks>
		int StartLazyDiscovering();
		/// <summary> Stops monitoring Lazy Discovering. </summary>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int StopLazyDiscovering();

		/// <summary> Query an IrDA IAS attribute value. </summary>
		/// <param name="Address"> An IrDA device's address. </param>
		/// <param name="ClassName"> The IAS attribute's class name. </param>
		/// <param name="AttributeName"> The IAS attribute's name. </param>
		/// <param name="Attribute"> On output contains the attribute's
		///   value. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		/// <seealso cref="wclIrDAAttribute" />
		int QueryAttribute(const unsigned long Address, const tstring& ClassName,
			const tstring& AttributeName, wclIrDAAttribute& Attribute);

		/// <summary> Subscribes for IrDA hardware changes notifications. </summary>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int Subscribe();
		/// <summary> Unsubscribes from IrDA hardware changes
		///   notifications. </summary>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int Unsubscribe();

		/// <summary> Gets the IrDA hardware status. </summary>
		/// <returns> <c>True</c> if the IrDA hardware is available. <c>False</c> if
		///   the IrDA hardware is not available. </returns>
		bool GetAvailable() const;
		/// <summary> Gets the IrDA hardware status. </summary>
		/// <value> <c>True</c> if the IrDA hardware is available. <c>False</c> if
		///   the IrDA hardware is not available. </value>
		__declspec(property(get = GetAvailable)) bool Available;

		/// <summary> Gets the Lazy Discovering state. </summary>
		/// <returns> <c>True</c> if Lazy Discovering is active.
		///   <c>False</c> otherwise. </returns>
		bool GetDiscovering() const;
		/// <summary> Gets the Lazy Discovering state. </summary>
		/// <value> <c>True</c> if Lazy Discovering is active.
		///   <c>False</c> otherwise. </value>
		__declspec(property(get = GetDiscovering)) bool Discovering;

		/// <summary> Gets the subscription state. </summary>
		/// <returns> <c>True</c> if the component is subscribed for the IrDA hardware
		///   changes notifications. <c>False</c> otherwise. </returns>
		bool GetSubscribed() const;
		/// <summary> Gets the subscription state. </summary>
		/// <value> <c>True</c> if the component is subscribed for the IrDA hardware
		///   changes notifications. <c>False</c> otherwise. </value>
		__declspec(property(get = GetSubscribed)) bool Subscribed;

		/// <summary> The event fires when the state of the IrDA hardware is
		///   changed. To check real status an application should read the
		///   <c>Available</c> property. </summary>
		/// <param name="Sender"> The object initiates the event. </param>
		/// <remarks> As there can be more than one IrDA adapter connected to PC
		///   removing (or plugging) any of them will fire the event. An application
		///   must use <c>Available</c> property to check if the IrDA subsystem
		///   is available or not (if there is at least one IrDA adapter
		///   connected). </remarks>
		wclNotifyEvent(OnChange);
		/// <summary> The even fires when new IrDA devices were found during
		///   Lazy Discovering. </summary>
		/// <param name="Sender"> The object initiates the event. </param>
		/// <param name="Device"> The record represents a just found IrDA
		///   device. </param>
		/// <seealso cref="wclIrDADevice" />
		wclIrDADevicesFoundEvent(OnDevicesFound);
	};
}