////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//   Wireless Communication Library 7                                         //
//                                                                            //
//   Copyright (C) 2006-2024 Mike Petrichenko                                 //
//                           Soft Service Company                             //
//                           All Rights Reserved                              //
//                                                                            //
//   http://www.btframework.com                                               //
//                                                                            //
//   support@btframework.com                                                  //
//   shop@btframework.com                                                     //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include "..\Communication\wclConnectionErrors.h"

namespace wclIrDA
{
	/* IrDA Error codes. */

	/// <summary> The base code for the IrDA errors. </summary>
	const int WCL_E_IRDA_BASE = 0x00040000;
	/// <summary> Access denied. </summary>
	const int WCL_E_IRDA_ACCESS_DENIED = WCL_E_IRDA_BASE + 0x0000;
	/// <summary> The initialization parameter is not a valid pointer or
	///   One of the option value or the option length parameters is not a valid
	///   part of the user address space, or the option length parameter is
	///   too small. </summary>
	const int WCL_E_IRDA_OPTION_IS_INVALID = WCL_E_IRDA_BASE + 0x0001;
	/// <summary> An invalid argument was supplied. </summary>
	const int WCL_E_IRDA_INVALID_ARGUMENT = WCL_E_IRDA_BASE + 0x0002;
	/// <summary> No more handles (descriptors) are available. </summary>
	const int WCL_E_IRDA_HANDLES_LIMIT_REACHED = WCL_E_IRDA_BASE + 0x0003;
	/// <summary> A blocking call is in progress, or the service provider is
	///   still processing a callback function. </summary>
	const int WCL_E_IRDA_IN_PROGRESS = WCL_E_IRDA_BASE + 0x0004;
	/// <summary> The descriptor is not a valid handle. </summary>
	const int WCL_E_IRDA_INVALID_HANDLE = WCL_E_IRDA_BASE + 0x0005;
	/// <summary> The specified protocol is the wrong type for this
	///   handle. </summary>
	const int WCL_E_IRDA_WRONG_PROTOCOL_TYPE = WCL_E_IRDA_BASE + 0x0006;
	/// <summary> The option is unknown or unsupported by the indicated protocol
	///   family (IrDA). </summary>
	const int WCL_E_IRDA_UNKNOW_OPTION = WCL_E_IRDA_BASE + 0x0007;
	/// <summary> The specified protocol is not supported. </summary>
	const int WCL_E_IRDA_PROTOCOL_NOT_SUPPORTED = WCL_E_IRDA_BASE + 0x0008;
	/// <summary> The specified handle type is not supported for IrDA
	///   communication. </summary>
	const int WCL_E_IRDA_HANDLE_NOT_SUPPORTED = WCL_E_IRDA_BASE + 0x0009;
	/// <summary> The IrDA is not supported. An infrared adapter and device
	///   driver is not installed on the local computer. </summary>
	const int WCL_E_IRDA_NOT_SUPPORTED = WCL_E_IRDA_BASE + 0x000A;
	/// <summary> Connection already exists. </summary>
	const int WCL_E_IRDA_CONNECTION_EXISTS = WCL_E_IRDA_BASE + 0x000B;
	/// <summary> An IrDA device or specific service was not found
	///   (address or service is invalid). </summary>
	const int WCL_E_IRDA_SERVICE_OR_DEVICE_NOT_FOUND = WCL_E_IRDA_BASE + 0x000C;
	/// <summary> The IrDA subsystem or the associated service provider has
	///   failed. </summary>
	const int WCL_E_IRDA_FAILED = WCL_E_IRDA_BASE + 0x000D;
	/// <summary> Connection has been terminated by a user. </summary>
	const int WCL_E_IRDA_CONNECTION_TERMINATED_BY_USER = WCL_E_IRDA_BASE + 0x000E;
	/// <summary> No buffer space is available. The handle cannot be
	///   created. </summary>
	const int WCL_E_IRDA_NO_BUFFERS = WCL_E_IRDA_BASE + 0x000F;
	/// <summary> Timeout occurred. </summary>
	const int WCL_E_IRDA_TIMEOUT = WCL_E_IRDA_BASE + 0x0010;
	/// <summary> The requested attribute was not found. </summary>
	const int WCL_E_IRDA_ATTRIBUTE_NOT_FOUND = WCL_E_IRDA_BASE + 0x0011;
	/// <summary> Connection has been rejected. </summary>
	const int WCL_E_IRDA_CONNECTION_REJECTED = WCL_E_IRDA_BASE + 0x0012;
	/// <summary> A limit on the number of tasks supported by the IrDA
	///   implementation has been reached. </summary>
	const int WCL_E_IRDA_LIMIT_REACHED = WCL_E_IRDA_BASE + 0x0013;
	/// <summary> The underlying subsystem is not ready for
	///   communication. </summary>
	const int WCL_E_IRDA_NOT_READY = WCL_E_IRDA_BASE + 0x0014;
	/// <summary> The version of IrDA subsystem support requested is not provided
	///   by this particular IrDA implementation. </summary>
	const int WCL_E_IRDA_VERSION_NOT_SUPPORTED = WCL_E_IRDA_BASE + 0x0015;
	/// <summary> A successful IrDA subsystem initialization must occur before
	///   using this function. </summary>
	const int WCL_E_IRDA_NOT_INITIALIZED = WCL_E_IRDA_BASE + 0x0016;
	/// <summary> The service provider returned an invalid or incomplete
	///   procedure table to the IrDA initialization. </summary>
	const int WCL_E_IRDA_INVALID_PROC_TABLE = WCL_E_IRDA_BASE + 0x0017;
	/// <summary> The service provider returned a version other
	///   than 2.2. </summary>
	const int WCL_E_IRDA_INVALID_PROVIDER = WCL_E_IRDA_BASE + 0x0018;
	/// <summary> The service provider failed to initialize. This error is
	///   returned if a layered service provider (LSP) or namespace provider was
	///   improperly installed or the provider fails to operate
	///   correctly. </summary>
	const int WCL_E_IRDA_SERVICE_PROFVIDER_FAILED = WCL_E_IRDA_BASE + 0x0019;
	/// <summary> Unexpected or unknown IrDA error. </summary>
	const int WCL_E_IRDA_UNEXPECTED = WCL_E_IRDA_BASE + 0x001A;
	/// <summary> Unable to create Lazy Discovering synchronization
	///   object. </summary>
	const int WCL_E_IRDA_UNABLE_CREATE_DISCOVERING_SYNC = WCL_E_IRDA_BASE + 0x001B;
	/// <summary> The requested attribute's class was not found. </summary>
	const int WCL_E_IRDA_ATTRIBUTE_CLASS_NOT_FOUND = WCL_E_IRDA_BASE + 0x001C;
	/// <summary> The Lazy Discovering is already running. </summary>
	const int WCL_E_IRDA_DISCOVERING_ACTIVE = WCL_E_IRDA_BASE + 0x001D;
	/// <summary> Unable to create Lazy Discovering thread. </summary>
	const int WCL_E_IRDA_UNABLE_CREATE_DISCOVERING_THREAD = WCL_E_IRDA_BASE + 0x001E;
	/// <summary> The Lazy Discovering is not running. </summary>
	const int WCL_E_IRDA_DISCOVERING_NOT_ACTIVE = WCL_E_IRDA_BASE + 0x001F;
	/// <summary> The component is already subscribed for the IrDA hardware change
	///   event. </summary>
	const int WCL_E_IRDA_ALREADY_SUBSCRIBED = WCL_E_IRDA_BASE + 0x0020;
	/// <summary> The component is not subscribed for the IrDA hardware change
	///   event. </summary>
	const int WCL_E_IRDA_NOT_SUBSCRIBED = WCL_E_IRDA_BASE + 0x0021;
	/// <summary> Client disconnected by remote device. </summary>
	const int WCL_E_IRDA_DISCONNECTED_BY_REMOTE_DEVICE = WCL_E_IRDA_BASE + 0x0022;
	/// <summary> Unable to create IrDA socket. </summary>
	const int WCL_E_IRDA_UNABLE_CREATE_SOCKET = WCL_E_IRDA_BASE + 0x0023;
	/// <summary> Unable to set a IrLMP communication mode. </summary>
	const int WCL_E_IRDA_UNABLE_SET_MODE = WCL_E_IRDA_BASE + 0x0024;
	/// <summary> Unable to create socket event. </summary>
	const int WCL_E_IRDA_UNABLE_CREATE_EVENT = WCL_E_IRDA_BASE + 0x0025;
	/// <summary> Unable to select socket events. </summary>
	const int WCL_E_IRDA_UNABLE_SET_EVENT = WCL_E_IRDA_BASE + 0x0026;
	/// <summary> Unable to get a read buffer size. </summary>
	const int WCL_E_IRDA_UNABLE_GET_READ_BUFFER = WCL_E_IRDA_BASE + 0x0027;
	/// <summary> Unable to get a write buffer size. </summary>
	const int WCL_E_IRDA_UNABLE_GET_WRITE_BUFFER = WCL_E_IRDA_BASE + 0x0028;
	/// <summary> Unable to set a read buffer size. </summary>
	const int WCL_E_IRDA_UNABLE_SET_READ_BUFFER = WCL_E_IRDA_BASE + 0x0029;
	/// <summary> Unable to set a write buffer size. </summary>
	const int WCL_E_IRDA_UNABLE_SET_WRITE_BUFFER = WCL_E_IRDA_BASE + 0x002A;
	/// <summary> Unable to resolve a remote IrDA device address. </summary>
	const int WCL_E_IRDA_UNABLE_TO_RESOLVE_ADDRESS = WCL_E_IRDA_BASE + 0x002B;
	/// <summary> This connection mode can not be used for IrDA server. </summary>
	const int WCL_E_IRDA_WRONG_CONNECTION_MODE = WCL_E_IRDA_BASE + 0x002C;
	/// <summary> Unable to create write completion event object. </summary>
	const int WCL_E_IRDA_UNABLE_TO_CREATE_WRITE_EVENT = WCL_E_IRDA_BASE + 0x002D;

	/* IrDA exceptions. */

	/// <summary> The base IrDA Framework exceptions class. </summary>
	/// <seealso cref="wclCommon::wclException" />
	class wclEIrDA : public wclException
	{
	public:
		/// <summary> Creates new exception object. </summary>
		/// <param name="msg"> The exception message. </param>
		wclEIrDA(const char* msg) : wclException(msg) { };
	};

	/// <summary> The exception used in IrDA client connection. </summary>
	/// <seealso cref="wclEIrDA" />
	class wclEIrDAClient : public wclEIrDA
	{
	public:
		/// <summary> Creates new exception object. </summary>
		/// <param name="msg"> The exception message. </param>
		wclEIrDAClient(const char* msg) : wclEIrDA(msg) { };
	};

	/// <summary> The exception used in IrDA server connection. </summary>
	/// <seealso cref="wclEIrDA" />
	class wclEIrDAServer : public wclEIrDA
	{
	public:
		/// <summary> Creates new exception object. </summary>
		/// <param name="msg"> The exception message. </param>
		wclEIrDAServer(const char* msg) : wclEIrDA(msg) { };
	};
}
