////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//   Wireless Communication Library 7                                         //
//                                                                            //
//   Copyright (C) 2006-2024 Mike Petrichenko                                 //
//                           Soft Service Company                             //
//                           All Rights Reserved                              //
//                                                                            //
//   http://www.btframework.com                                               //
//                                                                            //
//   support@btframework.com                                                  //
//   shop@btframework.com                                                     //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include "..\Common\wclErrors.h"

#include "..\Communication\wclConnections.h"

#include "wclIrDADevices.h"

using namespace wclCommon;
using namespace wclCommunication;

namespace wclIrDA
{
	/// <summary> The class handles server client connection. </summary>
	/// <remarks> An application must never create an object of this class
	///   directly. </remarks>
	/// <seealso cref="CwclServerClientDataConnection" />
	class CwclIrDAServerClientConnection : public CwclServerClientDataConnection
	{
		DISABLE_COPY(CwclIrDAServerClientConnection);

	private:
		unsigned long	FAddress;
		tstring			FName;

		WSAEVENT		FEvent;
		SOCKET			FSocket;
		int				FWriteResult;
		HANDLE			FWriteComplete;

		// We use IAS Query to resolve remote connected client name.
		tstring ResolveClientName() const;

	protected:
		/// <summary> Hardware-dependent method that prepares to communicate
		///   with a connected remote client device. </summary>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		virtual int HalAccept() override;
		/// <summary> Implements a hardware-dependent code that communicate with the
		///   connected remote device. </summary>
		/// <param name="Event"> The system even object handle (disconnect
		///   event). </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		virtual int HalCommunicate(const HANDLE Event) override;
		/// <summary> Implements a hardware-dependent code that disconnects from the
		///   connected remote device. </summary>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		virtual int HalDisconnect() override;

		/// <summary> Reads the read buffer size. </summary>
		/// <param name="Size"> On output he read buffer size in bytes. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		virtual int HalGetReadBufferSize(unsigned long& Size) override;
		/// <summary> Reads the write buffer size. </summary>
		/// <param name="Size"> On output the write buffer size in bytes. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		virtual int HalGetWriteBufferSize(unsigned long& Size) override;

		/// <summary> Sets the read buffer size. </summary>
		/// <param name="Size"> The new read buffer size in bytes. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		virtual int HalSetReadBufferSize(const unsigned long Size) override;
		/// <summary> Sets the write buffer size. </summary>
		/// <param name="Size"> The new write buffer size in bytes. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		virtual int HalSetWriteBufferSize(const unsigned long Size) override;

		/// <summary> Implements a hardware-dependent code that sends data to the
		///   connected device. </summary>
		/// <param name="Data"> The pointer to the data buffer that should be sent
		///   to the connected device. </param>
		/// <param name="Size"> The data buffer size. </param>
		/// <param name="Written"> The amount of bytes actually written to the
		///   device. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		virtual int HalWrite(const void* const Data, const unsigned long Size,
			unsigned long& Written) override;

	public:
		/// <summary> Creates new object of the server client connection. </summary>
		CwclIrDAServerClientConnection();

		/// <summary> Gets the connected IrDA client address. </summary>
		/// <returns> The IrDA  MAC address. </returns>
		unsigned long GetAddress() const;
		/// <summary> Gets the connected IrDA client address. </summary>
		/// <value> The IrDA  MAC address. </value>
		__declspec(property(get = GetAddress)) unsigned long Address;

		/// <summary> Gets the connected IrDA clients name. </summary>
		/// <returns> The IrDA clients's name. </returns>
		tstring GetName() const;
		/// <summary> Gets the connected IrDA clients name. </summary>
		/// <value> The IrDA clients's name. </value>
		__declspec(property(get = GetName)) tstring Name;
	};

	/// <summary> The prototype of the <c>OnConnect</c> event handler for IrDA
	///   server. </summary>
	/// <param name="Sender"> The object that initiated the event. </param>
	/// <param name="Client"> The
	///   <see cref="CwclIrDAServerClientConnection" /> represents a remote
	///   connected client. </param>
	/// <param name="Error"> The connection operation result code. If the
	///   parameter is <see cref="WCL_E_SUCCESS" /> the connection to a remote
	///   device was established and the remote device is connected. Otherwise
	///   the connection was not established and the remote device is
	///   not connected. </param>
	/// <seealso cref="CwclIrDAServerClientConnection" />
	#define wclIrDAServerConnectEvent(_event_name_) \
		__event void _event_name_(void* Sender, CwclIrDAServerClientConnection* const Client, \
		const int Error)
	/// <summary> A IrDA server connection <c>OnData</c> event handler
	///   prototype. </summary>
	/// <param name="Sender"> The object that initiated the event. </param>
	/// <param name="Client"> A
	///   <see cref="CwclIrDAServerClientConnection" /> object
	///   represents a remote client connection. </param>
	/// <param name="Data"> The pointer to the received data buffer.
	///   It is guaranteed that the parameter points to a valid data buffer.
	///   The data buffer is valid only inside the event handler. If an
	///   application needs to use the data outside the event handle it must
	///   allocate own buffer and copy data into it. </param>
	/// <param name="Size"> The data buffer size. It is guaranteed that the size
	///   is greater than 0. </param>
	/// <seealso cref="CwclIrDAServerClientConnection" />
	#define wclIrDAServerDataEvent(_event_name_) \
		__event void _event_name_(void* Sender, CwclIrDAServerClientConnection* const Client, \
		const void* const Data, const unsigned long Size)
	/// <summary> The prototype of the <c>OnDisconnect</c> event handler for IrDA server. </summary>
	/// <param name="Sender"> The object that initiated the event. </param>
	/// <param name="Client"> The
	///   <see cref="CwclIrDAServerClientConnection" /> represents a remote
	///   connected client. </param>
	/// <param name="Reason"> The disconnection code. If the
	///   parameter is <see cref="WCL_E_SUCCESS" /> the connection was closed
	///   by the <c>Disconnect</c> call. Otherwise indicates the disconnection
	///   reason. </param>
	/// <seealso cref="CwclIrDAServerClientConnection" />
	#define wclIrDAServerDisconnectEvent(_event_name_) \
		__event void _event_name_(void* Sender, CwclIrDAServerClientConnection* const Client, \
		const int Reason)

	/// <summary> The IrDA Server component. </summary>
	class CwclIrDAServer
	{
		DISABLE_COPY(CwclIrDAServer);

	private:
		wclIrDAConnectionMode	FMode;
		tstring					FService;

		CwclCustomConnection*	FConnection;
		
		void ServerClosed(void* Sender, const int Reason);
		void ServerConnect(void* Sender, CwclServerClientConnection* const Client,
			const int Error);
		void ServerData(void* Sender, CwclServerClientDataConnection* const Client,
			const void* const Data, const unsigned long Size);
		void ServerDisconnect(void* Sender, CwclServerClientConnection* const Client, 
			const int Reason);
		void ServerListen(void* Sender);

	protected:
		/// <summary> Fires the <c>OnClosed</c> event. </summary>
		/// <param name="Reason"> The close reason code.
		///   <see cref="WCL_E_SUCCESS" /> means that the server was closed by the
		///   <c>Close</c> method call. </param>
		virtual void DoClosed(const int Reason);
		/// <summary> Fires the <c>OnConnect</c> event. </summary>
		/// <param name="Client"> The
		///   <see cref="CwclIrDAServerClientConnection" /> represents a remote
		///   connected client. </param>
		/// <param name="Error"> The connection operation result code. If the
		///   parameter is <see cref="WCL_E_SUCCESS" /> the connection to a remote
		///   device was established and the remote device is connected. Otherwise
		///   the connection was not established and the remote device is
		///   not connected. </param>
		/// <seealso cref="CwclIrDAServerClientConnection" />
		virtual void DoConnect(CwclIrDAServerClientConnection* const Client, 
			const int Error);
		/// <summary> Fires the <c>OnCreateProcessor</c> event. </summary>
		/// <param name="Connection"> The
		///   <see cref="CwclServerClientDataConnection" /> object that requires a
		///   data processor. </param>
		/// <seealso cref="CwclServerClientDataConnection" />
		virtual void DoCreateProcessor(CwclServerClientDataConnection* const Connection);
		/// <summary> Fires the <c>OnData</c> event. </summary>
		/// <param name="Client"> A
		///   <see cref="CwclIrDAServerClientConnection" /> object
		///   represents a remote client connection. </param>
		/// <param name="Data"> The pointer to the received data buffer.
		///   It is guaranteed that the parameter points to a valid data buffer.
		///   The data buffer is valid only inside the event handler. If an
		///   application needs to use the data outside the event handle it must
		///   allocate own buffer and copy data into it. </param>
		/// <param name="Size"> The data buffer size. It is guaranteed that the size
		///   is greater than 0. </param>
		/// <seealso cref="CwclIrDAServerClientConnection" />
		virtual void DoData(CwclIrDAServerClientConnection* const Client, 
			const void* const Data, const unsigned long Size);
		/// <summary> Fires the <c>OnDestroyProcessor</c> event. </summary>
		/// <param name="Connection"> The
		///   <see cref="CwclServerClientDataConnection" /> object that requires a
		///   data processor. </param>
		/// <seealso cref="CwclServerClientDataConnection" />
		virtual void DoDestroyProcessor(CwclServerClientDataConnection* const Connection);
		/// <summary> Fires the <c>OnDisconnect</c> event. </summary>
		/// <param name="Client"> The
		///   <see cref="CwclIrDAServerClientConnection" /> represents a remote
		///   connected client. </param>
		/// <param name="Reason"> The disconnection code. If the
		///   parameter is <see cref="WCL_E_SUCCESS" /> the connection was closed
		///   by the <c>Disconnect</c> call. Otherwise indicates the disconnection
		///   reason. </param>
		/// <seealso cref="CwclIrDAServerClientConnection" />
		virtual void DoDisconnect(CwclIrDAServerClientConnection* const Client, 
			const int Reason);
		/// <summary> Fires the <c>OnListen</c> event. </summary>
		virtual void DoListen();

	public:
		/// <summary> Create a new IrDA server. </summary>
		CwclIrDAServer();
		/// <summary> Frees the component. </summary>
		virtual ~CwclIrDAServer();

		/// <summary> Closes the server. </summary>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int Close();
		/// <summary> Initializes the server and waits for incoming
		///   connections. </summary>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int Listen();

		/// <summary> The function adds the integer value into the server's IAS
		///   database. </summary>
		/// <param name="ClassName"> The class name for an IAS object. </param>
		/// <param name="AttribName"> The attribute name for an IAS object. </param>
		/// <param name="Value"> The 32-bit signed integer attribute
		///   value. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int AddIasRecord(const tstring& ClassName, const tstring& AttribName, 
			const int Value);
		/// <summary> The function adds the octet sequence value into the server's
		///   IAS database. </summary>
		/// <param name="ClassName"> The class name for an IAS object. </param>
		/// <param name="AttribName"> The attribute name for an IAS object. </param>
		/// <param name="Len"> The length of the bytes array pointed by the
		///   <c>Value</c> parameter. </param>
		/// <param name="Value"> The binary value (bytes array) to associate with
		///   the attribute. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		int AddIasRecord(const tstring& ClassName, const tstring& AttribName, 
			const unsigned short Len, const unsigned char* const Value);
		/// <summary> The function adds the string value into the server's
		///   IAS database. </summary>
		/// <param name="ClassName"> The class name for an IAS object. </param>
		/// <param name="AttribName"> The attribute name for an IAS object. </param>
		/// <param name="CharSet"> The string's character set. </param>
		/// <param name="Len"> The length in bytes of the string pointed by the
		///   <c>Value</c> parameter. </param>
		/// <param name="Value"> The user string to associate with the
		///   attribute. </param>
		/// <returns> If the function succeed the return value is
		///   <see cref="WCL_E_SUCCESS" />. Otherwise the method returns one of
		///   the WCL error codes. </returns>
		/// <seealso cref="wclIrDACharSet" />
		int AddIasRecord(const tstring& ClassName, const tstring& AttribName, 
			const wclIrDACharSet CharSet, const unsigned char Len, 
			const unsigned char* const Value);

		/// <summary> Gets a connected client by its index. </summary>
		/// <param name="Index"> Client's index. </param>
		/// <returns> The connected client. </returns>
		/// <seealso cref="CwclIrDAServerClientConnection" />
		CwclIrDAServerClientConnection* GetClients(const size_t Index) const;
		/// <summary> Gets a connected client by its index. </summary>
		/// <value> The connected client. </value>
		/// <seealso cref="CwclIrDAServerClientConnection" />
		__declspec(property(get = GetClients)) CwclIrDAServerClientConnection* Clients[];

		/// <summary> Gets the number of connected clients. </summary>
		/// <returns> The number of the connected clients. </returns>
		size_t GetClientsCount() const;
		/// <summary> Gets the number of connected clients. </summary>
		/// <value> The number of the connected clients. </value>
		__declspec(property(get = GetClientsCount)) size_t ClientsCount;

		/// <summary> Gets the server state. </summary>
		/// <returns> The <see cref="wclServerState" /> indicates the current
		///   connection state. </returns>
		/// <seealso cref="wclServerState" />
		wclServerState _GetState() const;
		/// <summary> Gets the server state. </summary>
		/// <value> The <see cref="wclServerState" /> indicates the current
		///   connection state. </value>
		/// <seealso cref="wclServerState" />
		__declspec(property(get = _GetState)) wclServerState State;

		/// <summary> Gets the IrDA connection mode. </summary>
		/// <returns> The IrDA connection mode. </returns>
		/// <seealso cref="wclIrDAConnectionMode" />
		wclIrDAConnectionMode GetMode() const;
		/// <summary> Sets the IrDA connection mode. </summary>
		/// <param name="Value"> The IrDA connection mode. </param>
		/// <remarks> If an application changes the property when server is
		///   listening the <see cref="wclEServerActive" /> exception raises.
		///   If an application sets mode to <c>cmIrLpt</c> the
		///   <see cref="wclEInvalidArgument" /> exception raises
		///   because that mode is not supported by the IrDA server. </remarks>
		/// <seealso cref="wclIrDAConnectionMode" />
		/// <exception cref="wclEServerActive"></exception>
		/// <exception cref="wclEInvalidArgument"></exception>
		void SetMode(const wclIrDAConnectionMode Value);
		/// <summary> Gets and sets the IrDA connection mode. </summary>
		/// <value> The IrDA connection mode. </value>
		/// <remarks> If an application changes the property when server is
		///   listening the <see cref="wclEServerActive" /> exception raises.
		///   If an application sets mode to <c>cmIrLpt</c> the
		///   <see cref="wclEInvalidArgument" /> exception raises
		///   because that mode is not supported by the IrDA server. </remarks>
		/// <seealso cref="wclIrDAConnectionMode" />
		/// <exception cref="wclEServerActive"></exception>
		/// <exception cref="wclEInvalidArgument"></exception>
		__declspec(property(get = GetMode, put = SetMode)) wclIrDAConnectionMode Mode;

		/// <summary> Gets the IrDA service that should be used to
		///   connect. It can be a TinyTP service name or LSAP selector number.
		///   In case of using LSAP selector number the format is 'LSAP-SELn'
		///   where <c>n</c> is a select number in range from 1 to 127. </summary>
		/// <returns> The IrDA service's name or LSAP selector number. </returns>
		tstring GetService() const;
		/// <summary> Sets the IrDA service that should be used to
		///   connect. It can be a TinyTP service name or LSAP selector number.
		///   In case of using LSAP selector number the format is 'LSAP-SELn'
		///   where <c>n</c> is a select number in range from 1 to 127. </summary>
		/// <param name="Value"> The IrDA service's name or LSAP selector number. </param>
		/// <remarks> If an application changes the property when server is
		///   listening the <see cref="wclEServerActive" /> exception raises. If an
		///   application sets the property to an empty string the
		///   <see cref="wclEInvalidArgument" /> exception raises. </remarks>
		/// <exception cref="wclEServerActive"></exception>
		/// <exception cref="wclEInvalidArgument"></exception>
		void SetService(const tstring& Value);
		/// <summary> Gets and sets the IrDA service that should be used to
		///   connect. It can be a TinyTP service name or LSAP selector number.
		///   In case of using LSAP selector number the format is 'LSAP-SELn'
		///   where <c>n</c> is a select number in range from 1 to 127. </summary>
		/// <value> The IrDA service's name or LSAP selector number. </value>
		/// <remarks> If an application changes the property when server is
		///   listening the <see cref="wclEServerActive" /> exception raises. If an
		///   application sets the property to an empty string the
		///   <see cref="wclEInvalidArgument" /> exception raises. </remarks>
		/// <exception cref="wclEServerActive"></exception>
		/// <exception cref="wclEInvalidArgument"></exception>
		__declspec(property(get = GetService, put = SetService)) tstring Service;

		/// <summary> The event fires when the server is closed. </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Reason"> The disconnection code. If the
		///   parameter is <see cref="WCL_E_SUCCESS" /> the connection was closed
		///   by the <c>Close</c> call. Otherwise indicates the disconnection
		///   reason. </param>
		wclClientConnectionDisconnectEvent(OnClosed);
		/// <summary> The event fires when a remote client is connected to
		///   the server. </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Client"> The
		///   <see cref="CwclIrDAServerClientConnection" /> represents a remote
		///   connected client. </param>
		/// <param name="Error"> The connection operation result code. If the
		///   parameter is <see cref="WCL_E_SUCCESS" /> the connection to a remote
		///   device was established and the remote device is connected. Otherwise
		///   the connection was not established and the remote device is
		///   not connected. </param>
		/// <seealso cref="CwclIrDAServerClientConnection" />
		wclIrDAServerConnectEvent(OnConnect);
		/// <summary> The event fires when connection to a remote device has
		///   been established with success to allow application provide a
		///   data processor for the connection. </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Connection"> The
		///   <see cref="CwclServerClientDataConnection" /> object that requires a
		///   Data Processor. </param>
		/// <remarks> If an application needs an additional data processing it
		///   can use Data Processors. An application must create a Data Processor
		///   inside this event handler. If a Data Processor created for the
		///   connection the <c>OnData</c> event will NOT fire. </remarks>
		/// <seealso cref="CwclCustomServerClientDataProcessor" />
		/// <seealso cref="CwclServerDataConnection" />
		/// <seealso cref="CwclServerClientDataConnection" />
		wclServerConnectionProcessorEvent(OnCreateProcessor);
		/// <summary> The event fires when new data is received from the
		///   connected remote device. </summary>
		/// <summary> A IrDA server connection <c>OnData</c> event handler
		///   prototype. </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Client"> A
		///   <see cref="CwclIrDAServerClientConnection" /> object
		///   represents a remote client connection. </param>
		/// <param name="Data"> The pointer to the received data buffer.
		///   It is guaranteed that the parameter points to a valid data buffer.
		///   The data buffer is valid only inside the event handler. If an
		///   application needs to use the data outside the event handle it must
		///   allocate own buffer and copy data into it. </param>
		/// <param name="Size"> The data buffer size. It is guaranteed that the size
		///   is greater than 0. </param>
		/// <seealso cref="CwclIrDAServerClientConnection" />
		wclIrDAServerDataEvent(OnData);
		/// <summary> The event fires when connection to a remote device has
		///   been terminated. An application must destroy the Data Processor
		///   created for the connection. </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Connection"> The
		///   <see cref="CwclServerClientDataConnection" /> object that requires a
		///   Data Processor. </param>
		/// <seealso cref="CwclCustomServerClientDataProcessor" />
		/// <seealso cref="CwclServerDataConnection" />
		/// <seealso cref="CwclServerClientDataConnection" />
		wclServerConnectionProcessorEvent(OnDestroyProcessor);
		/// <summary> The event fires when a remote client has disconnected from the
		///   server. </summary>
		/// <summary> The prototype of the <c>OnDisconnect</c> event handler for
		///   IrDA server. </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		/// <param name="Client"> The
		///   <see cref="CwclIrDAServerClientConnection" /> represents a remote
		///   connected client. </param>
		/// <param name="Reason"> The disconnection code. If the
		///   parameter is <see cref="WCL_E_SUCCESS" /> the connection was closed
		///   by the <c>Disconnect</c> call. Otherwise indicates the disconnection
		///   reason. </param>
		/// <seealso cref="CwclIrDAServerClientConnection" />
		wclIrDAServerDisconnectEvent(OnDisconnect);
		/// <summary> The event fires when the server is ready to accept incoming
		///   connections (is in listening state). </summary>
		/// <param name="Sender"> The object that initiated the event. </param>
		wclNotifyEvent(OnListen);
	};
}